"""Then system package fixtures for testinfra-bdd."""
from pytest_bdd import (
    then,
    parsers,
    when
)


@when(parsers.parse('the package is {package_name}'))
def the_package_is(package_name: str, testinfra_bdd_host):
    """
    Check the status of a package.

    Parameters
    ----------
    package_name : str
        The package name (e.g. "python3").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.package = testinfra_bdd_host.host.package(package_name.strip('"'))


@then(parsers.parse('the package state is {expected_status}'))
@then(parsers.parse('the package is {expected_status}'))
def the_package_status_is(expected_status, testinfra_bdd_host):
    """
    Check the status of a package (installed/absent).

    Parameters
    ----------
    expected_status : str
        Can be absent, installed or present.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the package is not in the expected state.
    """
    status_lookup = {
        'absent': False,
        'installed': True,
        'present': True
    }
    expected_to_be_installed = status_lookup[expected_status]
    pkg = testinfra_bdd_host.package
    actual_status = pkg.is_installed

    if expected_to_be_installed:
        message = f'Expected {pkg.name} to be {expected_status} on {testinfra_bdd_host.hostname} but it is absent.'
    else:
        message = f'Expected {pkg.name} to be absent on {testinfra_bdd_host.hostname} '
        message += 'but it is installed ({pkg.version}).'

    assert actual_status == expected_to_be_installed, message



"""Test exceptions are raised as expected."""
import pytest

from testinfra_bdd import get_host_fixture
from testinfra_bdd.address import when_the_address_and_port_is
from testinfra_bdd.command import the_command_is
from testinfra_bdd.pip import the_pip_package_is
from testinfra_bdd.pip import the_pip_package_state_is
from testinfra_bdd.process import the_process_filter_is


def test_invalid_command_stream_name():
    """Test that an exception is thrown when attempting to get an invalid resource type."""
    exception_raised = False

    try:
        host = get_host_fixture('docker://sut')
        the_command_is('ls', host)
        host.get_stream_from_command('foo')
    except ValueError as ex:
        exception_raised = True
        assert str(ex) == 'Unknown stream name "foo".'

    assert exception_raised, 'Expected an exception to be raised.'


def test_unready_host():
    """Test that a non-ready host throws an exception."""
    exception_raised = False

    try:
        get_host_fixture('docker://foo', 1)
    except AssertionError as ex:
        exception_raised = True
        assert str(ex) == 'The host docker://foo is not ready within 1 seconds.'

    assert exception_raised, 'Expected an exception to be raised.'


@pytest.mark.parametrize(
    'pip,expected_state,expected_exception_message',
    [
        (None, 'present', 'Pip package not set.  Have you missed a "When pip package is" step?'),
        ('semver', 'foo', 'Unknown expected state "foo" for a Pip package.'),
        ('semver', 'latest', 'Expected Pip package semver to be latest but it is superseded.')
    ]
)
def test_pip_package(pip, expected_state, expected_exception_message):
    """Test that a superseded pip package is identified."""
    exception_raised = False
    host = get_host_fixture('docker://sut')

    if pip:
        the_pip_package_is(pip, host)

    try:
        the_pip_package_state_is(expected_state, host)
    except (AssertionError, RuntimeError, ValueError) as ex:
        exception_raised = True
        assert str(ex) == expected_exception_message

    assert exception_raised, 'Expected an exception to be raised.'


@pytest.mark.parametrize(
    'process_specification',
    [
        '',
        'foo'
    ]
)
def test_invalid_process_specifications(process_specification):
    """Test that exceptions are raised when the process specification is invalid."""
    exception_raised = False
    expected_message = f'Unable to parse process filters "{process_specification}".'

    try:
        host = get_host_fixture('docker://sut')
        the_process_filter_is(process_specification, host)
    except ValueError as ex:
        exception_raised = True
        actual_message = str(ex)
        assert actual_message == expected_message

    assert exception_raised, 'Expected an exception to be raised.'


@pytest.mark.parametrize(
    'specification,exception_expected',
    [
        ('foo', True),
        ('foo:bar', True),
        ('localhost:0', False)
    ]
)
def test_invalid_addr_and_port_specifications(specification, exception_expected):
    """Test that exceptions are raised when the specification is invalid."""
    exception_raised = False
    expected_messages = {
        'foo': f'Unable to parse addr:port from "{specification}".',
        'foo:bar': f'Unable to parse addr:port from "{specification}". Unable to parse port.'
    }

    try:
        host = get_host_fixture('docker://sut')
        when_the_address_and_port_is(specification, host)
    except ValueError as ex:
        exception_raised = True
        actual_message = str(ex)
        assert actual_message == expected_messages[specification]

    if exception_expected:
        assert exception_raised, 'Expected an exception to be raised.'



"""The when steps of testinfra-bdd."""
import os
import pytest
from pytest_bdd import parsers
from pytest_bdd import when


@when(parsers.parse('the system property {property_name} is not "{expected_value}" skip tests'))
@when(parsers.parse('the system property {property_name} is not {expected_value} skip tests'))
def skip_tests_if_system_info_does_not_match(property_name, expected_value, testinfra_bdd_host):
    """
    Skip tests if a system property does not patch the expected value.

    Parameters
    ----------
    property_name : str
    expected_value : str
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    actual_value = testinfra_bdd_host.get_host_property(property_name)
    if actual_value != expected_value:
        pytest.skip(f'System {property_name} is {actual_value} which is not {expected_value}.')


@when(parsers.parse('the environment variable {key} is {value} skip tests'))
def skip_tests_if_env_key_is(key, value):
    """
    Skip tests if an environment variable is set to a particular value.

    Parameters
    ----------
    key : str
        The name of the environment variable.
    value : str
        The value the environment variable must be for the tests to be skipped.
    """
    if key in os.environ and os.environ[key] == value:
        pytest.skip(f'Environment variable {key} is set to {value}.')



"""
Then pip package fixtures for testinfra-bdd.

Please avoid already-imported warning: PYTEST_DONT_REWRITE.
"""
from pytest_bdd import (
    then,
    parsers,
    when
)

from testinfra_bdd.exception_message import exception_message


@when(parsers.parse('the pip package is {package_name}'))
def the_pip_package_is(package_name: str, testinfra_bdd_host):
    """
    Check the status of a pip package.

    Parameters
    ----------
    package_name : str
        The pip package name (e.g. "pytest-bdd").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.pip_package = testinfra_bdd_host.host.pip(package_name.strip('"'))


def check_entry_requirements(pip_package, expected_state):
    """
    Check that the entry requirements are met for the test.

    Parameters
    ----------
    pip_package : testinfra.Pip
        The Pip package to be checked.
    expected_state : str
        The expected state.

    Raises
    ------
    ValueError
        If the expected state is invalid.
    RuntimeError
        If the Pip package has not been set.
    """
    valid_expected_states = [
        'absent',
        'latest',
        'present',
        'superseded'
    ]

    if expected_state not in valid_expected_states:
        raise ValueError(f'Unknown expected state "{expected_state}" for a Pip package.')
    elif not pip_package:
        raise RuntimeError('Pip package not set.  Have you missed a "When pip package is" step?')


def get_pip_package_actual_state(pip_package, expected_state, host):
    """
    Get the actual state of a Pip package given the package and the expected state.

    Parameters
    ----------
    pip_package : testinfra.Pip
        The Pip package to be checked.
    expected_state : str
        The expected state.
    host : testinfra.host.Host
        The host to be checked against.

    Returns
    -------
    tuple
        str
            The actual state (e.g. absent, latest, present or superseded).
        str
            A suitable message if the actual state doesn't match the actual state.
    """
    state_checks = [
        'absent',
        'present'
    ]

    check_entry_requirements(pip_package, expected_state)

    if expected_state in state_checks:
        actual_state = 'absent'

        if pip_package.is_installed:
            actual_state = 'present'

        return actual_state, exception_message(
            f'Pip package {pip_package.name}',
            actual_state,
            expected_state
        )

    outdated_packages = host.pip.get_outdated_packages()

    if pip_package.name in outdated_packages:
        actual_state = 'superseded'
    else:
        actual_state = 'latest'

    return actual_state, exception_message(f'Pip package {pip_package.name}', actual_state, expected_state)


@then('the pip check is OK')
def the_pip_check_is_ok(testinfra_bdd_host):
    """
    Verify installed packages have compatible dependencies.

    Parameters
    ----------
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the packages have incompatible dependencies.
    """
    host = testinfra_bdd_host.host
    cmd = host.pip.check()
    message = f'Incompatible Pip packages - {cmd.stdout} {cmd.stderr}'
    assert cmd.rc == 0, message


@then(parsers.parse('the pip package state is {expected_state}'))
@then(parsers.parse('the pip package is {expected_state}'))
def the_pip_package_state_is(expected_state, testinfra_bdd_host):
    """
    Check the state of a Pip package.

    Parameters
    ----------
    expected_state : str
        The expected state of the package.  Can be absent, latest or installed.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the actual state doesn't match the expected state.
    """
    (actual_state, message) = get_pip_package_actual_state(
        testinfra_bdd_host.pip_package,
        expected_state,
        testinfra_bdd_host.host
    )
    assert actual_state == expected_state, message


@then(parsers.parse('the pip package version is {expected_version}'))
def the_pip_package_version_is(expected_version, testinfra_bdd_host):
    """
    Check the version of a Pip package.

    Parameters
    ----------
    expected_version : str
        The version of the package that is expected.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the actual version is not the expected version.
    """
    pip_package = testinfra_bdd_host.pip_package
    assert pip_package, 'Pip package not set.  Have you missed a "When pip package is" step?'
    actual_version = pip_package.version
    message = f'Expected Pip package version to be {expected_version} but it was {actual_version}.'
    assert actual_version == expected_version, message



"""The main fixture for the testinfra-bdd tests."""
import time

import testinfra


class TestinfraBDD:
    """A class that is used as the fixture in the given/when/then steps."""

    def __init__(self, url):
        """
        Create a TestinfraBDD object.

        Initialises the host attribute.

        Parameters
        ----------
        url : str
            The URL of the System Under Test (SUT).  Must comply to the Testinfra
            URL patterns.  See https://testinfra.readthedocs.io/en/latest/backends.html
        """
        self.address = None
        self.arch = None
        self.codename = None
        self.command = None
        self.distribution = None
        self.file = None
        self.group = None
        self.host = testinfra.get_host(url)
        self.hostname = None
        self.package = None
        self.pip_package = None
        self.port = None
        self.port_number = None
        self.process_specification = None
        self.processes = None
        self.release = None
        self.service = None
        self.socket = None
        self.socket_url = None
        self.type = None
        self.url = url
        self.user = None

    def get_host_property(self, property_name):
        """
        Get a named host property.

        Parameters
        ----------
        property_name : str
            The name of the property to be extracted.

        Raises
        ------
        AssertError
            If the property_name is invalid.

        Returns
        -------
        str
            The value of the property.
        """
        properties = {
            'type': self.host.system_info.type,
            'distribution': self.host.system_info.distribution,
            'release': self.host.system_info.release,
            'codename': self.host.system_info.codename,
            'arch': self.host.system_info.arch,
            'hostname': self.host.backend.get_hostname(),
            'connection_type': self.host.backend.NAME
        }

        assert property_name in properties, f'Invalid host property name "{property_name}".'
        return properties[property_name]

    def get_stream_from_command(self, stream_name):
        """
        Get a named stream from the command.

        Parameters
        ----------
        stream_name : str
            The name of the stream.

        Raises
        ------
        AssertError
            If the command attribute is None.
        ValueError
            When the stream name is not recognized.

        Returns
        -------
        str
            The requested stream content.
        """
        assert self.command, 'No command has been executed.'

        if stream_name == 'stdout':
            return self.command.stdout
        elif stream_name == 'stderr':
            return self.command.stderr

        raise ValueError(f'Unknown stream name "{stream_name}".')

    def is_host_ready(self, timeout=0):
        """
        Check if a host is ready within a specified time.

        Will poll the host every second until timeout number of seconds have
        expired.  If this host has not responded within that time, the host
        is assumed to not be ready.

        Parameters
        ----------
        timeout : int,optional
            The time in seconds to wait for the host to become ready.  The
            default is zero.

        Returns
        -------
        bool
            True if the host is responding to the host.system_info.type request.
            False if it doesn't.
        """
        if timeout:
            self.wait_until_is_host_ready(timeout)

        try:
            self.host.system_info.type
            self.arch = self.host.system_info.arch
            self.codename = self.host.system_info.codename
            self.distribution = self.host.system_info.distribution
            self.hostname = self.host.backend.hostname
            self.release = self.host.system_info.release
            self.type = self.host.system_info.type
            is_ready = True
        except AssertionError:
            is_ready = False

        return is_ready

    def wait_until_is_host_ready(self, timeout=0):
        """
        Check if a host is ready within a specified time.

        Will poll the host every second until timeout number of seconds have
        expired.  If this host has not responded within that time, the host
        is assumed to not be ready.

        Parameters
        ----------
        timeout : int,optional
            The time in seconds to wait for the host to become ready.  The
            default is zero.

        Returns
        -------
        bool
            True if the host is responding to the host.system_info.type request.
            False if it doesn't.
        """
        is_ready = False
        now = time.time()
        deadline = now + timeout

        while now <= deadline and not is_ready:
            is_ready = self.is_host_ready()

            if now < deadline and not is_ready:
                time.sleep(1)

            now = time.time()

        return is_ready



"""
Then address fixtures for testinfra-bdd.

Please avoid already-imported warning: PYTEST_DONT_REWRITE.
"""
from pytest_bdd import (
    then,
    parsers,
    when
)

from testinfra_bdd.parsers import parse_addr_and_port


@when(parsers.parse('the address is {address}'))
def when_the_address_is(address: str, testinfra_bdd_host):
    """
    Check the status of a user.

    Parameters
    ----------
    username : str
        The address (e.g. "www.ggoogle.com").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.address = testinfra_bdd_host.host.addr(address.strip('"'))


@when(parsers.parse('the address and port is {url}'))
def when_the_address_and_port_is(url, testinfra_bdd_host):
    """
    Check the status of an address and port.

    Parameters
    ----------
    url: str
        The hostname and port (e.g. "www.google.com:443").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture
    """
    (
        testinfra_bdd_host.address,
        testinfra_bdd_host.port,
        testinfra_bdd_host.port_number
    ) = parse_addr_and_port(url.strip('"'), testinfra_bdd_host.host)


@then(parsers.parse('the address is {expected_state}'))
def the_address_is(expected_state, testinfra_bdd_host):
    """
    Check the actual state of an address against an expected state.

    Parameters
    ----------
    expected_state : str
        The expected state of the address.

    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual state does not match the state.
    """
    address = testinfra_bdd_host.address
    assert address, 'Address is not set.  Did you miss a "When address is" step?'
    properties = {
        'resolvable': address.is_resolvable,
        'reachable': address.is_reachable
    }
    expected_state = expected_state.strip('"')
    assert expected_state in properties, f'Invalid state for {address.name} ("{expected_state}").'
    message = f'Expected the address {address.name} to be {expected_state} but it is not.'
    assert properties[expected_state], message


@then(parsers.parse('the port is {expected_state}'))
def the_port_is(expected_state, testinfra_bdd_host):
    """
    Check the actual state of an address port against an expected state.

    Parameters
    ----------
    expected_state : str
        The expected state of the port.

    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual state does not match the state.
    """
    port = testinfra_bdd_host.port
    expected_state = expected_state.strip('"')
    assert port, 'Port is not set.  Did you miss a "When the address and port" step?'
    properties = {
        'reachable': port.is_reachable
    }
    assert expected_state in properties, f'Unknown Port property ("{expected_state}").'
    message = f'{testinfra_bdd_host.address.name}:{testinfra_bdd_host.port_number} is unreachable.'
    assert properties['reachable'], message



"""The given steps of testinfra-bdd."""
import testinfra_bdd.fixture
from pytest_bdd import given
from pytest_bdd import parsers


@given(parsers.parse('the host with URL "{hostspec}" is ready'), target_fixture='testinfra_bdd_host')
def the_host_is_ready(hostspec):
    """
    Ensure that the host is ready within the specified number of seconds.

    If the host does not become ready within the specified number of seconds,
    fail the tests.

    Parameters
    ----------
    hostspec : str
        The URL of the System Under Test (SUT).  Must comply to the Testinfra
        URL patterns.  See
        https://testinfra.readthedocs.io/en/latest/backends.html

    Returns
    -------
    testinfra_bdd.fixture.TestinfraBDD
        The object to return as a fixture.
    """
    return testinfra_bdd.get_host_fixture(hostspec)


@given(parsers.parse('the host with URL "{hostspec}" is ready within {seconds:d} seconds'),
       target_fixture='testinfra_bdd_host')
def the_host_is_ready_with_a_number_of_seconds(hostspec, seconds):
    """
    Ensure that the host is ready within the specified number of seconds.

    If the host does not become ready within the specified number of seconds,
    fail the tests.

    Parameters
    ----------
    hostspec : str
        The URL of the System Under Test (SUT).  Must comply to the Testinfra
        URL patterns.  See
        https://testinfra.readthedocs.io/en/latest/backends.html
    seconds : int
        The number of seconds that the host is expected to become ready in.

    Returns
    -------
    testinfra_bdd.fixture.TestinfraBDD
        The object to return as a fixture.
    """
    return testinfra_bdd.get_host_fixture(hostspec, seconds)



"""
Then command fixtures for testinfra-bdd.

Please avoid already-imported warning: PYTEST_DONT_REWRITE.
"""
import re

from pytest_bdd import parsers
from pytest_bdd import then
from pytest_bdd import when


@when(parsers.parse('the command is {command}'))
def the_command_is(command: str, testinfra_bdd_host):
    """
    Execute and check the status of a command.

    Parameters
    ----------
    command : str
        The command (e.g. "ls -l /etc/motd").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.command = testinfra_bdd_host.host.run(command.strip('"'))


@then(parsers.parse('the command {command} exists in path'))
@then(parsers.parse('the command "{command}" exists in path'))
def check_command_exists_in_path(command, testinfra_bdd_host):
    """
    Assert that a specified command is present on the host path.

    Parameters
    ----------
    command : str
        The name of the command to check for.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the command is not found on the path.
    """
    message = f'Unable to find the command "{command}" on the path.'
    assert testinfra_bdd_host.host.exists(command.strip('"')), message


@then(parsers.parse('the command {stream_name} contains "{text}"'))
def check_command_stream_contains(stream_name, text, testinfra_bdd_host):
    """
    Check that the stdout or stderr stream contains a string.

    Parameters
    ----------
    stream_name : str
        The name of the stream to check.  Must be "stdout" or "stderr".
    text : str
        The text to search for.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the specified stream does not contain the expected text.
    """
    stream = testinfra_bdd_host.get_stream_from_command(stream_name)
    message = f'The string "{text}" was not found in the {stream_name} ("{stream}") of the command.'
    assert text in stream, message


@then(parsers.parse('the command {stream_name} contains the expected value'))
def the_command_stderr_contains_the_expected_value(stream_name, expected_value, testinfra_bdd_host):
    """
    Check that the stdout or stderr stream contains a pre-defined expected value.

    Parameters
    ----------
    stream_name : str
        The name of the stream to check.  Must be "stdout" or "stderr".
    expected_value : A pytest fixture.
        This is pre-defined by the user.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the specified stream does not contain the expected text.
    """
    stream = testinfra_bdd_host.get_stream_from_command(stream_name)
    message = f'The string "{expected_value}" was not found in the {stream_name} ("{stream}") of the command.'
    assert expected_value in stream, message


@then(parsers.parse('the command {stream_name} does not contain "{text}"'))
def the_command_stdout_does_not_contain_foo(stream_name, text, testinfra_bdd_host):
    """
    Check that the stdout or stderr stream does not contain a string.

    Parameters
    ----------
    stream_name : str
        The name of the stream to check.  Must be "stdout" or "stderr".
    text : str
        The text to search for.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the specified stream does contain the unexpected text.
    """
    stream = testinfra_bdd_host.get_stream_from_command(stream_name)
    message = f'The unexpected string "{text}" was found in the {stream_name} ("{stream}") of the command.'
    assert text not in stream, message


@then(parsers.parse('the command {stream_name} contains the regex "{pattern}"'))
def check_command_stream_contains_the_regex(stream_name, pattern, testinfra_bdd_host):
    """
    Check that the stdout or stderr stream matches a regular expression pattern.

    Parameters
    ----------
    stream_name : str
        The name of the stream to be checked.  Must be stdout or stderr.
    pattern : str
        The pattern to search for in the stream.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the specified stream does not match the pattern.
    ValueError
        When the stream name is not recognized.
    """
    stream = testinfra_bdd_host.get_stream_from_command(stream_name)
    message = f'The regex "{pattern}" is not found in the {stream_name} "{stream}".'
    # The parsers.parse function escapes the parsed string.  We need to clean it up before using it.
    pattern = pattern.encode('utf-8').decode('unicode_escape')
    prog = re.compile(pattern)
    assert prog.search(stream) is not None, message


@then(parsers.parse('the command return code is {expected_return_code:d}'))
def check_command_return_code(expected_return_code, testinfra_bdd_host):
    """
    Check that the expected return code from a command matches the actual return code.

    Parameters
    ----------
    expected_return_code : int
        The expected return code (e.g. zero/0).
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the actual return code does not match the expected return code.
    """
    cmd = testinfra_bdd_host.command
    actual_return_code = cmd.rc
    message = f'Expected a return code of {expected_return_code} but got {actual_return_code}.'
    assert expected_return_code == actual_return_code, message


@then(parsers.parse('the command {stream_name} is empty'))
def command_stream_is_empty(stream_name, testinfra_bdd_host):
    """
    Check that the specified command stream is empty.

    Parameters
    ----------
    stream_name : str
        The name of the stream to be checked.  Must be stdout or stderr.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the specified stream does not match the pattern.
    """
    stream = testinfra_bdd_host.get_stream_from_command(stream_name)
    assert not stream, f'Expected {stream_name} to be empty ("{stream}").'



"""Then socket fixtures for testinfra-bdd."""
from pytest_bdd import (
    then,
    parsers,
    when
)


@when(parsers.parse('the socket is {socket}'))
def when_the_socket_is(socket, testinfra_bdd_host):
    """
    Check the status of a socket.

    Parameters
    ----------
    socket : str
        URL of the socket (e.g. "tpc://22").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.socket = testinfra_bdd_host.host.socket(socket.strip('"'))


@then(parsers.parse('the socket is {expected_state}'))
def the_socket_is(expected_state, testinfra_bdd_host):
    """
    Check the state of a socket.

    Parameters
    ----------
    expected_state : str
        The expected state of the socket.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual state does not match the state.
    """
    socket = testinfra_bdd_host.socket
    socket_url = testinfra_bdd_host.socket_url
    actual_state = 'not listening'
    assert socket, 'Socket is not set.  Have you missed a "When socket is" step?'

    if socket.is_listening:
        actual_state = 'listening'

    message = f'Expected socket {socket_url} to be {expected_state} but it is {actual_state}.'
    assert actual_state == expected_state, message



"""
Then process fixtures for testinfra-bdd.

Please avoid already-imported warning: PYTEST_DONT_REWRITE.
"""
from pytest_bdd import (
    then,
    parsers,
    when
)

from testinfra_bdd.parsers import parse_process_filters


@when(parsers.parse('the process filter is {process_specification}'))
def the_process_filter_is(process_specification, testinfra_bdd_host):
    """
    Check the status of processes.

    Parameters
    ----------
    process_specification : str
        The process specification (e.g. "user=root,comm=ntpd").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    process_specification = process_specification.strip('"')
    testinfra_bdd_host.process_specification = process_specification
    host = testinfra_bdd_host.host
    testinfra_bdd_host.processes = host.process.filter(**parse_process_filters(process_specification))


@then(parsers.parse('the process count is {expected_count:d}'))
def the_process_count_is(expected_count, testinfra_bdd_host):
    """
    Check that the process count matches the expected count.

    Parameters
    ----------
    expected_count : int
        The expected number of processes.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual process count does not match the expected count.
    """
    specification = testinfra_bdd_host.process_specification
    processes = testinfra_bdd_host.processes
    assert processes, 'No process set, did you forget a "When process filter" step?'
    actual_process_count = len(processes)
    message = f'Expected process specification "{specification}" to return {expected_count} '
    message += f'but found {actual_process_count} "{processes}".'
    assert actual_process_count == expected_count, message



"""Basic string parsers for Testinfra BDD."""


def parse_addr_and_port(addr_and_port, host):
    """
    Parse a string containing an address and port.

    Parameters
    ----------
    addr_and_port : str
        The address and port in the format of "host:port".
    host : testinfra.host
        The relevant host.

    Returns
    -------
    tuple
        str the address, int the port number.

    Raises
    ------
    ValueError
        If the string can't be parsed.
    """
    strings = addr_and_port.split(':')
    message = f'Unable to parse addr:port from "{addr_and_port}".'

    if len(strings) != 2:
        raise ValueError(message)

    addr = strings[0]

    try:
        port_number = int(strings[1])
    except ValueError:
        raise ValueError(message + ' Unable to parse port.')

    address = host.addr(addr)
    port = address.port(port_number)
    return address, port, port_number


def parse_process_filters(specification):
    """
    Parse the process filters into a dictionary.

    Parameters
    ----------
    specification : str
        The process specifications to be parsed.

    Returns
    -------
    dict
        A dictionary of the parsed filters.

    Raises
    ------
    ValueError
        If the specification can't be parsed.
    """
    filters = {}

    for keypair in specification.split(','):
        keypair = keypair.split('=')

        if len(keypair) != 2:
            raise ValueError(f'Unable to parse process filters "{specification}".')

        key = keypair[0]
        value = keypair[1]
        filters[key] = value

    return filters



"""Then user fixtures for testinfra-bdd."""
from pytest_bdd import (
    then,
    parsers,
    when
)


@when(parsers.parse('the user is {username}'))
def the_user_is(username: str, testinfra_bdd_host):
    """
    Check the status of a user.

    Parameters
    ----------
    username : str
        The user name (e.g. "ben").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.user = testinfra_bdd_host.host.user(username.strip('"'))


@then(parsers.parse('the user {property_name} is {expected_value}'))
def the_user_property_is(property_name, expected_value, testinfra_bdd_host):
    """
    Check the property of a user.

    Parameters
    ----------
    property_name : str
        The name of the property to compare.
    expected_value : str
        The value that is expected.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual value does not match the expected value.
    """
    user = testinfra_bdd_host.user
    assert user, 'User not set.  Have you missed a "When user is" step?'

    if testinfra_bdd_host.user.exists:
        actual_state = 'present'
        properties = {
            'gid': str(user.gid),
            'group': user.group,
            'home': user.home,
            'shell': user.shell,
            'state': actual_state,
            'uid': str(user.uid)
        }
    else:
        actual_state = 'absent'
        properties = {
            'state': actual_state
        }

    assert property_name in properties, f'Unknown user property "{property_name}".'
    actual_value = properties[property_name]
    message = f'Expected {property_name} for user {user.name} to be "{expected_value}" '
    message += f'but it was "{actual_value}".'
    assert actual_value == expected_value, message


@then(parsers.parse('the user is {expected_state}'))
def check_the_user_state(expected_state, testinfra_bdd_host):
    """
    Check that the actual state of a user matches the expected state.

    Parameters
    ----------
    expected_state : str
        The expected state (e.g. absent or present).
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    the_user_property_is('state', expected_state, testinfra_bdd_host)



"""Then file fixtures for testinfra-bdd."""
from pytest_bdd import (
    then,
    parsers,
    when
)


@when(parsers.parse('the group is {groupname}'))
def the_group_is(groupname: str, testinfra_bdd_host):
    """
    Check the status of a group.

    Parameters
    ----------
    groupname : str
        The group name (e.g. "admin").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.group = testinfra_bdd_host.host.group(groupname.strip('"'))


@then(parsers.parse('the group {property_name} is {expected_value}'))
def the_group_property_is(property_name, expected_value, testinfra_bdd_host):
    """
    Check the property of a group.

    Parameters
    ----------
    property_name : str
        The name of the property to compare.
    expected_value : str
        The value that is expected.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual value does not match the expected value.
    """
    group = testinfra_bdd_host.group
    assert group, 'Group not set.  Have you missed a "When group is" step?'

    properties = {
        'gid': None,
        'state': 'absent'
    }

    if group.exists:
        properties = {
            'gid': str(group.gid),
            'state': 'present'
        }

    assert property_name in properties, f'Unknown group property ({property_name}).'
    actual_value = properties[property_name]
    message = f'Expected group property to be {expected_value} but it was {actual_value}.'
    assert actual_value == expected_value, message


@then(parsers.parse('the group is {expected_state}'))
def check_the_group_state(expected_state, testinfra_bdd_host):
    """
    Check that the actual state of a group matches the expected state.

    Parameters
    ----------
    expected_state : str
        The expected state (e.g. absent or present).
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    the_group_property_is('state', expected_state, testinfra_bdd_host)



"""Dynamic configuration for setuptools."""
import os
import setuptools

try:
    import testinfra_bdd
    version = testinfra_bdd.__version__
except ModuleNotFoundError:
    version = os.environ['VERSION']

with open('README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()


install_requires = []

with open('pyproject.toml') as stream:
    pyproject = stream.readlines()

for line in pyproject:
    line = line.strip()[1:-2]

    if line.startswith('jmespath>') or line.startswith('pytest-bdd>') or line.startswith('pytest-testinfra>'):
        install_requires.append(line)

setuptools.setup(
    name='testinfra-bdd',
    version=version,
    author='Cloud Based DQ Ltd.',
    author_email='info@cbdq.io',
    description='An interface between pytest-bdd and pytest-testinfra.',
    install_requires=install_requires,
    keywords='testinfra,bdd',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/cbdq-io/testinfra-bdd',
    project_urls={
        'Bug Tracker': 'https://github.com/cbdq-io/testinfra-bdd/issues',
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ],
    package_dir={'': '.'},
    packages=setuptools.find_packages(where='.'),
    python_requires='>=3.6')



"""
An interface between pytest-bdd and pytest-testinfra.

For documentation and examples, please go to
https://github.com/cbdq-io/testinfra-bdd

Please avoid already-imported warning: PYTEST_DONT_REWRITE.
"""
from testinfra_bdd.fixture import TestinfraBDD

"""PYTEST_MODULES.

A list of all testinfra-bdd packages that contain fixtures.
"""
PYTEST_MODULES = [
    'testinfra_bdd',
    'testinfra_bdd.given',
    'testinfra_bdd.address',
    'testinfra_bdd.command',
    'testinfra_bdd.file',
    'testinfra_bdd.group',
    'testinfra_bdd.package',
    'testinfra_bdd.pip',
    'testinfra_bdd.process',
    'testinfra_bdd.service',
    'testinfra_bdd.socket',
    'testinfra_bdd.user',
    'testinfra_bdd.when'
]

"""The version of the module.

This is used by setuptools and by gitchangelog to identify the name of the name
of the release.
"""
__version__ = '2.2.5'


def get_host_fixture(hostspec, timeout=0):
    """
    Return a host that is confirmed as ready.

    hostspec : str
        The URL of the System Under Test (SUT).  Must comply to the Testinfra
        URL patterns.  See
        https://testinfra.readthedocs.io/en/latest/backends.html
    timeout : int, optional
        The number of seconds that the host is expected to become ready in.

    Returns
    -------
    testinfra_bdd.fixture.TestinfraBDD
        The object to return as a fixture.

    Raises
    ------
    AssertError
        When the host is not ready.
    """
    if timeout:
        message = f'The host {hostspec} is not ready within {timeout} seconds.'
    else:
        message = f'The host {hostspec} is not ready.'

    host = TestinfraBDD(hostspec)
    assert host.is_host_ready(timeout), message
    return host



"""Then service fixtures for testinfra-bdd."""
from pytest_bdd import parsers
from pytest_bdd import then
from pytest_bdd import when


@when(parsers.parse('the service is {service}'))
def the_service_is(service: str, testinfra_bdd_host):
    """
    Check the status of a service.

    Parameters
    ----------
    service : str
        The service name (e.g. "named").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.service = testinfra_bdd_host.host.service(service.strip('"'))


@then('the service is not enabled')
def the_service_is_not_enabled(testinfra_bdd_host):
    """
    Check that the service is not enabled.

    Parameters
    ----------
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the service is enabled.
    """
    service = testinfra_bdd_host.service
    message = f'Expected {service.name} on host {testinfra_bdd_host.hostname} to be disabled, but it is enabled.'
    assert not service.is_enabled, message


@then('the service is enabled')
def the_service_is_enabled(testinfra_bdd_host):
    """
    Check that the service is enabled.

    Parameters
    ----------
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the service is not enabled.
    """
    service = testinfra_bdd_host.service
    message = f'Expected {service.name} on host {testinfra_bdd_host.hostname} to be enabled, but it is disabled.'
    assert service.is_enabled, message


@then('the service is not running')
def the_service_is_not_running(testinfra_bdd_host):
    """
    Check that the service is not running.

    Parameters
    ----------
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the service is running.
    """
    service = testinfra_bdd_host.service
    message = f'Expected {service.name} on host {testinfra_bdd_host.hostname} to not be running.'
    assert not service.is_running, message


@then('the service is running')
def the_service_is_running(testinfra_bdd_host):
    """
    Check that the service is running.

    Parameters
    ----------
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the service is not running.
    """
    service = testinfra_bdd_host.service
    message = f'Expected {service.name} on host {testinfra_bdd_host.hostname} to be running.'
    assert service.is_running, message



"""Then file fixtures for testinfra-bdd."""
import jmespath
import json
import re

from pytest_bdd import (
    then,
    parsers,
    when
)

from testinfra_bdd.exception_message import exception_message


@when(parsers.parse('the file is {file_name}'))
def the_file_is(file_name: str, testinfra_bdd_host):
    """
    Check the status of a file.

    Parameters
    ----------
    file_name : str
        The package name (e.g. "/etc/motd").
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.
    """
    testinfra_bdd_host.file = testinfra_bdd_host.host.file(file_name.strip('"'))


@then(parsers.parse('the file contents contains "{text}"'))
def the_file_contents_contains_text(text, testinfra_bdd_host):
    """
    Check if the file contains a string.

    Parameters
    ----------
    text : str
        The string to search for in the file content.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the file does not contain the string.
    """
    file = testinfra_bdd_host.file
    assert file.contains(text), f'The file {testinfra_bdd_host.hostname}:{file.path} does not contain "{text}".'


@then(parsers.parse('the file contents contains the regex "{pattern}"'))
def the_file_contents_matches_the_regex(pattern, testinfra_bdd_host):
    """
    Check if the file contains matches a regex pattern.

    Parameters
    ----------
    pattern : str
        The regular expression to match against the file content.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the regex does not match the file content.
    """
    file = testinfra_bdd_host.file
    file_name = f'{testinfra_bdd_host.hostname}:{file.path}'
    message = f'The regex "{pattern}" does not match the content of {file_name} ("{file.content_string}").'
    # The parsers.parse function escapes the parsed string.  We need to clean it up before using it.
    pattern = pattern.encode('utf-8').decode('unicode_escape')
    assert re.search(pattern, file.content_string) is not None, message


@then(parsers.parse('the file is {expected_status}'))
def the_file_status(expected_status, testinfra_bdd_host):
    """
    Check if the file is present or absent.

    Parameters
    ----------
    expected_status : str
        Should be present or absent.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        When the expected status does not match the actual status.
    """
    the_file_property_is('state', expected_status, testinfra_bdd_host)


@then(parsers.parse('the file {property_name} is {expected_value}'))
def the_file_property_is(property_name, expected_value, testinfra_bdd_host):
    """
    Check the property of a file.

    Parameters
    ----------
    property_name : str
        The name of the property to compare.
    expected_value : str
        The value that is expected.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the actual value does not match the expected value.
    """
    (actual_value, exception_message) = get_file_actual_state(
        testinfra_bdd_host.file,
        property_name,
        expected_value
    )
    assert actual_value == expected_value, exception_message


@then(parsers.parse('the JMESPath expression {expression} returns {expected_value}'))
def the_jmespath_expression_expression_returns_expected_value(expression, expected_value, testinfra_bdd_host):
    """
    Check the contents of a JSON file with JMESPath.

    Parameters
    ----------
    expression : str
        A JMESPath expression.
    expected_value : str
        The value expected to be returned.  All values returned by JMESPath will
        be converted to a string before comparison.
    testinfra_bdd_host : testinfra_bdd.fixture.TestinfraBDD
        The test fixture.

    Raises
    ------
    AssertError
        If the file is not valid JSON or if the JMESPath expression returns
        another value other than the expected value.
    """
    file = testinfra_bdd_host.file
    file_name = f'{testinfra_bdd_host.hostname}:{file.path}'
    the_file_property_is('state', 'present', testinfra_bdd_host)
    the_file_property_is('type', 'file', testinfra_bdd_host)
    data = json.loads(file.content_string)
    actual_value = str(jmespath.search(expression, data))
    message = f'Expected {expression} in {file_name} to be "{expected_value}", but it is "{actual_value}".'
    assert actual_value == expected_value, message


def get_file_actual_state(file, property_name, expected_state):
    """
    Get the actual state of a file given the package and the expected state.

    Parameters
    ----------
    file : testinfra.File
        The file to be checked.
    property_name : str
        The name of the property to check (e.g. state).
    expected_state : str
        The expected state.

    Returns
    -------
    tuple
        str
            The actual state (e.g. absent, latest, present or superseded).
        str
            A suitable message if the actual state doesn't match the actual state.
    """
    properties = get_file_properties(file)
    assert property_name in properties, f'Unknown user property "{property_name}".'
    actual_state = properties[property_name]
    return actual_state, exception_message(f'File {file.path} {property_name}', actual_state, expected_state)


def get_file_properties(file):
    """
    Get the properties of the file.

    Parameters
    ----------
    file : testinfra.File
        The file to be checked.

    Returns
    -------
    dict
        A dictionary of the properties.
    """
    assert file, 'File not set.  Have you missed a "When file is" step?'
    properties = {
        'group': None,
        'mode': None,
        'owner': None,
        'state': 'absent',
        'type': None,
        'user': None
    }

    if file.exists:
        properties = {
            'group': file.group,
            'mode': '0o%o' % file.mode,
            'owner': file.user,
            'state': 'present',
            'type': get_file_type(file),
            'user': file.user
        }

    return properties


def get_file_type(file):
    """
    Get the file type.

    Parameters
    ----------
    file : testinfra.File
        The file to be checked.

    Returns
    -------
    str
        The type of file.
    """
    file_type = None

    type_lookup = {
        'file': file.is_file,
        'directory': file.is_directory,
        'pipe': file.is_pipe,
        'socket': file.is_socket,
        'symlink': file.is_symlink
    }

    for key in type_lookup:
        if type_lookup[key]:
            file_type = key
            break

    return file_type
